/* 
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 *
 */

package org.apache.muse.tools.generator.synthesizer;

import java.io.File;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;

import javax.xml.namespace.QName;

import org.apache.muse.tools.generator.util.Capability;
import org.apache.muse.tools.inspector.JavaMethod;
import org.apache.muse.tools.inspector.JavaProperty;
import org.apache.muse.util.ReflectUtils;

public abstract class AbstractSynthesizer implements Synthesizer {
	static final String INDENT = "    ";

	static final String REQUEST_SUFFIX = "Request";
	
	private static final String JAVA_CLASS_NAME = "MyCapability";

	static final String INTERFACE_PREFIX = "I";
	
	private int _prefixCounter = 0;

	private HashMap _prefixes = new HashMap();

	protected String createFileName(String interfaceName) {
		return interfaceName.replaceAll("\\.", "\\" + File.separator) + ".java";
	}

	protected String generateClassName(Capability capability) {
		return generateClassName(null, capability);
	}

	protected String generateClassName(String prefix, Capability capability) {
		String implementingClass = capability.getImplementingClass();
		
		String packageName = null;
		String shortClassName = null;
		
		if(implementingClass == null) {
			packageName = ClassInfo.getPackageName(capability.getURI());
			shortClassName = JAVA_CLASS_NAME;
		} else {
			packageName = ReflectUtils.getPackageName(implementingClass);
			shortClassName = ReflectUtils.getShortName(implementingClass);
		}
		
		if(prefix != null) {
			shortClassName = prefix + shortClassName;
		}
		
		return packageName + "." + shortClassName;
	}
	
	protected String getMethodName(JavaMethod method) {
		String name = method.getJavaName();
		if (name.endsWith(REQUEST_SUFFIX)) {
			name = name.substring(0, name.indexOf(REQUEST_SUFFIX));
		}
		return name;
	}
	
	protected void generateClassDef(String remoteClassName, boolean isInterface, StringBuffer code) {
		generateClassDef(remoteClassName, null, null, isInterface, code);
	}
	
	protected void generateClassDef(String remoteClassName, String extendsClass, String[] implementsClasses ,boolean isInterface, StringBuffer code) {
		code.append("public ");
		code.append(isInterface?"interface":"class");
		code.append(" ");
		code.append(ReflectUtils.getShortName(remoteClassName));
		
		if(extendsClass != null) {
			code.append(" extends " + extendsClass);			
		}
		
		if(implementsClasses != null && implementsClasses.length > 0) {
			code.append(" implements ");
			code.append(implementsClasses[0]);
			
			for(int i=1; i < implementsClasses.length; i++) {				
				code.append(", " + implementsClasses[i]);				
			}
		}
		
		newLine(code);
		generateOpenBlock(code);
		newLine(code);
	}
	
	protected void generateImports(ClassInfo classInfo, StringBuffer code) {
		Iterator i = classInfo.getImports().iterator();

		while (i.hasNext()) {
			String className = needsImport((Class) i.next());
			if (className != null) {
				statement("import " + className + ";", code);
				newLine(code);
			}
		}

		newLine(code);
	}

	protected String needsImport(Class className) {
		if (className.isPrimitive()) {
			return null;
		}
		if (className.isArray()) {
			className = ReflectUtils.getClassFromArrayClass(className);
		}
		if (className.getName().startsWith("java.lang.")) {
			return null;
		}
		return className.getName();
	}

	protected String convertType(Class returnType, ClassInfo classInfo) {
		boolean isArray = returnType.isArray();

		returnType = isArray ? ReflectUtils.getClassFromArrayClass(returnType)
				: returnType;

		// Use fully-qualified name if conflict exists
		if (classInfo.hasConflict(returnType)) {
			return returnType.getName();
		}
		return ReflectUtils.getShortName(returnType) + (isArray ? "[]" : "");
	}

	protected String getPropertyName(JavaProperty property, boolean forSetter) {
		String name = property.getName().getLocalPart();
		if (!forSetter) {
			return name;
		}
		name = name.substring(0, 1).toUpperCase() + name.substring(1);
		return name;
	}

	protected String getParamName(QName paramName, int position) {
		String name;

		if (paramName == null) {
			name = "param" + position;
		} else {
			name = paramName.getLocalPart();
		}

		return name;
	}

	protected void statement(String statement, StringBuffer code) {
		code.append(statement);
	}

	protected void generateHeaderComment(String className, StringBuffer code) {
		String shortName = ReflectUtils.getShortName(className);
		String fileName = shortName + ".java";

		comment(code);
		newLine(code);

		comment(fileName, code);
		newLine(code);

		comment(new Date().toString(), code);
		newLine(code);

		comment("Generated by the Apache Muse Code Generation Tool", code);
		newLine(code);

		comment(code);
		newLine(code);
	}
	
	protected void generatePackage(String className, StringBuffer code) {
		String packageName = ReflectUtils.getPackageName(className);

		if (packageName != null) {
			statement("package " + packageName + ";", code);
			newLine(2, code);
		}

	}

	protected void comment(StringBuffer code) {
		comment("", code);
	}

	protected void comment(String comment, StringBuffer code) {
		code.append("// " + comment);
	}

	protected void newLine(StringBuffer code) {
		newLine(1, code);
	}

	protected void newLine(int n, StringBuffer code) {
		for (int i = 0; i < n; i++) {
			code.append("\n");
		}
	}

	protected void generateCloseBlock(StringBuffer code) {
		code.append("}");
	}

	protected void generateOpenBlock(StringBuffer code) {
		code.append("{");
	}

	protected void indent(StringBuffer code) {
		indent(1, code);
	}

	protected void indent(int indent, StringBuffer code) {
		for (int i = 0; i < indent; i++) {
			code.append(INDENT);
		}
	}

	protected String getObjectName(String name, Class type) {
		if (!type.isPrimitive())
			return name;

		else if (type == boolean.class)
			return "new Boolean(" + name + ')';

		else if (type == double.class)
			return "new Double(" + name + ')';

		else if (type == float.class)
			return "new Float(" + name + ')';

		else if (type == int.class)
			return "new Integer(" + name + ')';

		else if (type == long.class)
			return "new Long(" + name + ')';

		return "new Short(" + name + ')';
	}

	protected String getPrimitiveClassName(Class type) {
		if (type == boolean.class)
			return "Boolean";

		else if (type == float.class || type == double.class)
			return "Float";

		return "Integer";
	}

	protected String getPrimitiveConvertMethodName(String type) {
		if (type.endsWith("Boolean"))
			return "booleanValue";

		else if (type.endsWith("Float") || type.endsWith("Double"))
			return "floatValue";

		return "intValue";
	}

	protected String getArrayValue(Class type) {
		if (!type.isPrimitive())
			return "value";

		if (type == boolean.class)
			return "Boolean.toString(value)";

		else if (type == float.class || type == double.class)
			return "Double.toString(value)";

		return "Long.toString(value)";
	}

	protected String getNullValue(Class type) {
		if (!type.isPrimitive())
			return "null";

		if (type == boolean.class)
			return "false";

		return "0";
	}
	
	protected void generateQName(QName qname, StringBuffer code) {
		generateQName(qname.getNamespaceURI(), qname.getLocalPart(), code);
	}

	protected void generateQName(String uri, String name, StringBuffer code) {
		code.append("new QName(\"");

		if (uri != null && uri.length() > 0) {
			code.append(uri);
			code.append("\", \"");
		}

		code.append(name);
		code.append("\", \"");

		code.append(getPrefix(uri));
		code.append("\")");
	}

	protected String getPrefix(String uri) {
		String prefix = (String) _prefixes.get(uri);

		if (prefix == null) {
			prefix = "pfx" + _prefixCounter++;
			_prefixes.put(uri, prefix);
		}

		return prefix;
	}
	
	protected StringBuffer beginHeaderCode(String className) {
		StringBuffer code = new StringBuffer();
		
		generateHeaderComment(className, code);
		generatePackage(className, code);

		return code;
	}
}
