/* 
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 *
 */

package org.apache.muse.tools.generator.synthesizer;

import java.lang.reflect.Constructor;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;

import javax.xml.namespace.QName;

import org.apache.muse.core.proxy.ProxyHandler;
import org.apache.muse.core.proxy.ReflectionProxyHandler;
import org.apache.muse.tools.generator.util.ConfigurationData;
import org.apache.muse.tools.generator.util.ConfigurationDataDescriptor;
import org.apache.muse.tools.inspector.JavaMethod;
import org.apache.muse.tools.inspector.JavaProperty;
import org.apache.muse.tools.inspector.ResourceInspector;
import org.apache.muse.util.ReflectUtils;
import org.apache.muse.util.xml.XsdUtils;
import org.w3c.dom.Document;

public class ProxyClassSynthesizer extends ProxyInterfaceSynthesizer {
	static ConfigurationDataDescriptor[] REQUIRED_PARAMETERS = 
		new ConfigurationDataDescriptor[] {
			ConfigurationData.CAPABILITIES_MAP_LIST_CONFIGURATION,
			ConfigurationData.GENERATE_CUSTOM_HEADERS_CONFIGURATION,
			ConfigurationData.WSDL_DOCUMENT_LIST_CONFIGURATION
		};
	
	StringBuffer _operationNamesCode;
	StringBuffer _returnTypesCode;
	StringBuffer _actionsCode;
	
	StringBuffer _requestNamesCode;
	StringBuffer _responseNamesCode;
	StringBuffer _requestParamNamesCode;
	
	StringBuffer _propertiesOperationsCode;
	private String _interfaceName;

	private boolean _hasMethods = false;
	private boolean _hasProperties = false;
	
	protected String generateClassName(Document wsdlDocument) {
		_interfaceName = super.generateClassName(wsdlDocument);
		return _interfaceName + "Proxy";
	}

	protected void initializeCode(String className) {
		_importSet = new HashSet();
		addImport(_baseClientClass);
		
        _className = className;
		_headerCode = beginHeaderCode(_className);
        _operationsCode = beginOperationsCode();
        
        _operationNamesCode = beginOperationNamesCode();
        _returnTypesCode = beginReturnTypesCode();
        _actionsCode = beginActionsCode();
        
        _requestNamesCode = beginRequestNamesCode();
        _responseNamesCode = beginResponseNamesCode();
        _requestParamNamesCode = beginRequestParamNamesCode();
        
        resetIndex();
	}
	
	private StringBuffer beginRequestParamNamesCode() {
		StringBuffer code = new StringBuffer();
		
		indent(code);
		code.append("private static final QName[][] _REQUEST_PARAM_NAMES = {");
		newLine(code);
		_importSet.add(QName.class);
		return code;
	}

	private StringBuffer beginResponseNamesCode() {
		StringBuffer code = new StringBuffer();
		
		indent(code);
		code.append("private static final QName[] _RESPONSE_NAMES = {");
		newLine(code);
		indent(2,code);
		_importSet.add(QName.class);
		return code;
	}

	private StringBuffer beginRequestNamesCode() {
		StringBuffer code = new StringBuffer();
		
		indent(code);
		code.append("private static final QName[] _REQUEST_NAMES = {");
		newLine(code);
		indent(2,code);
		_importSet.add(QName.class);
		return code;
	}

	private StringBuffer beginActionsCode() {
		StringBuffer code = new StringBuffer();
		
		indent(code);
		code.append("private static final String[] _ACTIONS = {");
		newLine(code);
		indent(2,code);
		return code;
	}

	private StringBuffer beginReturnTypesCode() {
		StringBuffer code = new StringBuffer();
		
		indent(code);
		code.append("private static final Class[] _RETURN_TYPES = {");
		newLine(code);
		indent(2,code);
		return code;
	}

	private StringBuffer beginOperationsCode() {
		return new StringBuffer();
	}

	private StringBuffer beginOperationNamesCode() {
		StringBuffer code = new StringBuffer();
		
		indent(code);
		code.append("private static final String[] _METHOD_NAMES = {");
		newLine(code);
		indent(2,code);
		return code;
	}
	
	protected void updateCode(ClassInfo classInfo) {
		updateMethods(classInfo, _operationsCode);		
		classInfo.addImports(_importSet);
	}
	
	protected void generateMethodQualifier(StringBuffer code) {
		code.append("public ");
	}
	
	protected void generateMethodBody(JavaMethod method, ClassInfo classInfo, StringBuffer code) {
		_hasMethods = true;
		
		newLine(code);
		indent(code);
		generateOpenBlock(code);
		newLine(code);

		generateParamConversion(method, code);
		generateHandler(method, classInfo, code);
		generateInvoke(method, code);

		indent(code);
		generateCloseBlock(code);
		
		updateMethodNames(method);
		updateReturnTypes(method);
		updateActions(method);
		
		updateRequestNames(method);
		updateResponseNames(method);
		updateRequestParamNamesCode(method);
	}
	
	private void generateParamConversion(JavaMethod operation, StringBuffer code) {
		QName[] paramNames = operation.getParameterTypeNames();
		Class[] paramTypes = operation.getParameterTypes();

		indent(2, code);
		code.append("Object[] params = new Object[");
		code.append(paramTypes.length);
		code.append("];");
		newLine(code);

		if (paramTypes.length > 0)
			newLine(code);

		for (int n = 0; n < paramTypes.length; ++n) {
			indent(2, code);
			code.append("params[");
			code.append(n);
			code.append("] = ");

			String name = ResourceInspector.getLowerCamelName(paramNames[n]
					.getLocalPart());
			code.append(getObjectName(name, paramTypes[n]));
			code.append(";");
			newLine(code);
		}

		newLine(code);
	}

	private void generateHandler(JavaMethod method, ClassInfo classInfo, StringBuffer code) {
		indent(2, code);
		code.append("ProxyHandler handler = getHandler(\"");
		code.append(method.getJavaName());
		code.append("\");");
		newLine(code);
		
		classInfo.addImport(ProxyHandler.class);
	}
	
	private void generateInvoke(JavaMethod method, StringBuffer code) {
		indent(2, code);

		Class returnType = method.getReturnType();

		if (returnType == void.class) {
			code.append(getInvokeString() + ";");
		}
		else if (!returnType.isPrimitive()) {
			code.append("return (");
			code.append(ReflectUtils.getShortName(returnType));
			code.append(")" + getInvokeString() + ";");
			
		} else {
			String className = getPrimitiveClassName(returnType);
			String convertMethodName = getPrimitiveConvertMethodName(className);

			code.append(className);
			code.append(" result = (");
			code.append(className);
			code.append(")" + getInvokeString() + ";");
			newLine(code);

			indent(2, code);
			code.append("return result.");
			code.append(convertMethodName);
			code.append("();");
		}
		
		newLine(code);
	}
	
	private String getInvokeString() {
		if(_generateCustomHeaders) {
			return "invoke(handler, params, customHeaders)";
		}

		return "invoke(handler, params)";
	}

	private void updateRequestParamNamesCode(JavaMethod method) {
		QName[] names = method.getParameterTypeNames();
		
		indent(2, _requestParamNamesCode);
		generateOpenBlock(_requestParamNamesCode);
		newLine(_requestParamNamesCode);
		
		for(int i=0; i < names.length; i++) {
			indent(3, _requestParamNamesCode);
			generateQName(names[i], _requestParamNamesCode);
			_requestParamNamesCode.append(",");
			newLine(_requestParamNamesCode);
		}
				
		if(names.length > 0) {
			int length = _requestParamNamesCode.length();
			_requestParamNamesCode.delete(length - 2, length);
		}
		
		newLine(_requestParamNamesCode);
		indent(2, _requestParamNamesCode);
		generateCloseBlock(_requestParamNamesCode);
		_requestParamNamesCode.append(",");
		newLine(_requestParamNamesCode);
	}

	private void updateResponseNames(JavaMethod method) {
		QName returnName = method.getReturnName();
		if (returnName != null && !returnName.equals(XsdUtils.ANY_TYPE_QNAME))
			generateQName(returnName, _responseNamesCode);
        
        else
        	_responseNamesCode.append("null");
		
		_responseNamesCode.append(", ");
	}

	private void updateRequestNames(JavaMethod method) {
		QName returnName = method.getName();
		if (returnName != null && !returnName.equals(XsdUtils.ANY_TYPE_QNAME))
			generateQName(returnName, _requestNamesCode);
        
        else
        	_requestNamesCode.append("null");
		
		_requestNamesCode.append(", ");
	}
	
	private void updateMethodNames(JavaMethod method) {
		_operationNamesCode.append('"');
		_operationNamesCode.append(method.getJavaName());
		_operationNamesCode.append('"');
		_operationNamesCode.append(", ");
	}
	
	private void updateReturnTypes(JavaMethod method) {
		Class type = method.getReturnType();
		
		if(type.isArray()) {
			type = ReflectUtils.getClassFromArrayClass(type);
		}
		
		_returnTypesCode.append(ReflectUtils.getShortName(type));
		_returnTypesCode.append(".class");
		_returnTypesCode.append(", ");
	}
	
	private void updateActions(JavaMethod method) {
		_actionsCode.append('"');
		_actionsCode.append(method.getActionURI());
		_actionsCode.append('"');
		_actionsCode.append(", ");
	}

	protected void generatePropertyDeleteBody(JavaProperty property, ClassInfo classInfo, int propertyIndex, StringBuffer code) {		
		_hasProperties = true;
		
		newLine(code);
		indent(code);
		generateOpenBlock(code);
		newLine(code);

		indent(2, code);
		code.append("deleteResourceProperty(PROPERTIES[");
		code.append(propertyIndex);
		code.append("]);\n");

		indent(code);
		generateCloseBlock(code);
	}

	protected void generatePropertyGetBody(JavaProperty property, ClassInfo classInfo, int propertyIndex, StringBuffer code) {
		_hasProperties = true;
		
		Class type = property.getJavaType();
		newLine(code);
		indent(code);
		generateOpenBlock(code);
		newLine(code);

		if (type.isArray()) {
			indent(2, code);
			code.append("return (");
			code.append(ReflectUtils.getShortName(type));
			code.append(")getPropertyAsObject(");
			code.append("PROPERTIES[");
			code.append(propertyIndex);
			code.append("], ");

			type = ReflectUtils.getClassFromArrayClass(type);
			code.append(ReflectUtils.getShortName(type));
			code.append(".class);");
			newLine(code);
		}

		else {
			indent(2, code);

			Class array = ReflectUtils.getArrayClassFromClass(type);
			code.append(ReflectUtils.getShortName(array));

			code.append(" results = (");
			code.append(ReflectUtils.getShortName(array));
			code.append(")getPropertyAsObject(");
			code.append("PROPERTIES[");
			code.append(propertyIndex);
			code.append("], ");
			code.append(ReflectUtils.getShortName(type));
			code.append(".class);");
			newLine(code);
			indent(2, code);
			code.append("return results.length == 0 ? ");
			code.append(getNullValue(type));
			code.append(" : results[0];");
			newLine(code);
		}

		indent(code);
		generateCloseBlock(code);
	}

	protected void generatePropertySetBody(JavaProperty property, ClassInfo classInfo, int propertyIndex, String setType, StringBuffer code) {
		_hasProperties = true;
		
		Class type = property.getJavaType();
		if(type.isArray()) {
			type = ReflectUtils.getClassFromArrayClass(type);
		}
		
		newLine(code);
		indent(code);
		generateOpenBlock(code);
		newLine(code);
		indent(2, code);

		code.append(setType);
		code.append("ResourceProperty(PROPERTIES[");
		code.append(propertyIndex);
		code.append("], ");

		if (type.isArray())
			code.append("value");

		else {
			code.append("new Object[]{ ");
			code.append(getArrayValue(type));
			code.append(" }");
		}

		code.append(");");
		newLine(code);

		indent(code);
		generateCloseBlock(code);		
	}
	
	public String generateCombinedCode(ClassInfo classInfo) {			
		endDeclarationCode(_operationNamesCode);
		endDeclarationCode(_returnTypesCode);
		endDeclarationCode(_actionsCode);
		
		endDeclarationCode(_requestNamesCode);
		endDeclarationCode(_responseNamesCode);
		endDeclarationCode(_requestParamNamesCode);
		
		StringBuffer footerCode = generateFooterCode();
		
		classInfo.addImports(_importSet);
		endHeaderCode(classInfo);
		
		StringBuffer code = new StringBuffer();
		
		code.append(_headerCode);

		if(_hasProperties || _hasMethods) {
			code.append(_operationsCode);     
			newLine(code);
		}
		
		if(_hasMethods) {			
        	code.append(_operationNamesCode);
        	newLine(code);
        	code.append(_returnTypesCode);
        	newLine(code);
        	code.append(_actionsCode);
        	newLine(code);
        	code.append(_requestNamesCode);
        	newLine(code);
        	code.append(_responseNamesCode);
        	newLine(code);
        	code.append(_requestParamNamesCode);
		}
        
		code.append(footerCode);
		
		return code.toString();
	}
	
	protected void endHeaderCode(ClassInfo classInfo) {
		generateImports(classInfo, _headerCode);
		String[] interfaces = { ReflectUtils.getShortName(_interfaceName) };
		String parentClass = ReflectUtils.getShortName(_baseClientClass);
		generateClassDef(_className, parentClass, interfaces, false, _headerCode);
	}
	
	private StringBuffer generateFooterCode() {
		StringBuffer code = new StringBuffer();
		
		generateProxyConstructors(code);

		if(_hasMethods) {		
			generateGetHandler(code);
			
			newLine(code);
			generateStatic(code);					
		}
		
		generateCloseBlock(code);	
		
		newLine(code);
		return code;
	}
	
	private void generateProxyConstructors(StringBuffer code) {		
		String className = ReflectUtils.getShortName(_className);

		Constructor[] ctors = _baseClientClass.getConstructors();

		for (int n = 0; n < ctors.length; ++n) {
			indent(code);
			code.append("public ");
			code.append(className);

			Class[] params = ctors[n].getParameterTypes();
			generateParamList(null, ctors[n].getParameterTypes(), code);
			addImports(ctors[n].getParameterTypes());

			indent(code);
			generateOpenBlock(code);
			newLine(code);
			
			indent(2,code);
			code.append("super(");

			for (int i = 0; i < params.length; ++i) {
				code.append(PARAM_NAME + i);

				if (i != params.length - 1)
					code.append(", ");
			}

			code.append(");");
			newLine(code);
			
			indent(code);
			generateCloseBlock(code);
			newLine(2, code);
		}
	}

	private void generateGetHandler(StringBuffer code) {
		indent(code);
		code.append("protected ProxyHandler getHandler(String methodName)");
		addImport(ProxyHandler.class);
		newLine(code);

		indent(code);
		generateOpenBlock(code);
		newLine(code);

		indent(2, code);
		statement("return (ProxyHandler)_HANDLERS_BY_NAME.get(methodName);",
				code);
		newLine(code);

		indent(code);
		generateCloseBlock(code);
	}
	
	private void generateStatic(StringBuffer code) {
		
		newLine(code);
		indent(code);
		statement(
				"private static final Map _HANDLERS_BY_NAME = new HashMap();",
				code);
		
		addImport(Map.class);
		addImport(HashMap.class);
		
		newLine(2, code);
		indent(code);
		code.append("static");
		
		newLine(code);
		indent(code);
		generateOpenBlock(code);
		newLine(code);

		indent(2,code);
		code.append("for (int n = 0; n < _METHOD_NAMES.length; ++n)");
		
		newLine(code);
		indent(2,code);
		generateOpenBlock(code);
		
		newLine(code);
		indent(3,code);
		code.append("ProxyHandler handler = new ReflectionProxyHandler();");
		addImport(ReflectionProxyHandler.class);
		
		newLine(code);
		indent(3,code);
		code.append("handler.setAction(_ACTIONS[n]);");
		
		newLine(code);
		indent(3,code);
		code.append("handler.setRequestName(_REQUEST_NAMES[n]);");
		
		newLine(code);
		indent(3,code);
		code.append("handler.setRequestParameterNames(_REQUEST_PARAM_NAMES[n]);");
		
		newLine(code);
		indent(3,code);
		code.append("handler.setResponseName(_RESPONSE_NAMES[n]);");
		
		newLine(code);
		indent(3,code);
		code.append("handler.setReturnType(_RETURN_TYPES[n]);");

		newLine(2,code);
		indent(3,code);
		code
				.append("_HANDLERS_BY_NAME.put(_METHOD_NAMES[n], handler);");
		
		newLine(code);
		indent(2,code);
		generateCloseBlock(code);
		newLine(code);
		
		indent(code);
		generateCloseBlock(code);
	}		
	
	private void endDeclarationCode(StringBuffer code) {
    	int length = code.length();
    	code.delete(length - 2, length);
        
    	newLine(code);
    	indent(code);
    	generateCloseBlock(code);
    	code.append(";");
    	newLine(code);   
	}	

	public ConfigurationDataDescriptor[] getConfigurationDataDescriptions() {
		return REQUIRED_PARAMETERS;
	}
}
