/* 
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 *
 */

package org.apache.muse.tools.generator.synthesizer;

import java.io.File;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import javax.xml.namespace.QName;

import org.apache.muse.tools.generator.util.Capability;
import org.apache.muse.tools.inspector.JavaMethod;
import org.apache.muse.tools.inspector.JavaProperty;
import org.apache.muse.util.ReflectUtils;
import org.apache.muse.util.messages.Messages;
import org.apache.muse.util.messages.MessagesFactory;
import org.apache.muse.ws.addressing.soap.SoapFault;

public class ServerClassSynthesizer extends ServerInterfaceSynthesizer {

	private static Messages _MESSAGES = MessagesFactory.get(ServerClassSynthesizer.class);
	
	protected void generateCapability(Capability capability, Map files, Set ignoreSet) {
		ClassInfo classInfo = new ClassInfo(capability);
		String className = generateClassName(capability);
		initializeCode(className, classInfo);
		
		addImport(getBaseClass(capability));
		
		if(classInfo.needsInitializer()) {
			generateInitialize(classInfo, _bodyCode);
		}
				
		if(!capability.isEmpty()) {
			generatePropertiesDeclarations(classInfo, _bodyCode);
			generatePropertyOperations(classInfo, _bodyCode);
			generateOperations(classInfo, _bodyCode);
		}
		
		classInfo.addImports(_importSet);
		capability.setImplementingClass(className);
		String classFileName = createFileName(className);		
		files.put(classFileName, generateCombinedCode(classInfo));
		ignoreSet.add(new File(classFileName));
	}
	
	protected void endHeaderCode(ClassInfo classInfo) {
		generateImports(classInfo, _headerCode);
		Capability capability = classInfo.getCapability();	
		String[] interfaces = null;
		if(!capability.isEmpty()) {
			interfaces = new String[]{ ReflectUtils.getShortName(generateClassName(INTERFACE_PREFIX,capability)) };
		}
		String parentClass = convertType(getBaseClass(capability), classInfo);
		generateClassDef(_className, parentClass, interfaces , false, _headerCode);	
	}
	
	protected void generatePropertiesDeclarations(ClassInfo classInfo, StringBuffer code) {
		Capability capability = classInfo.getCapability();
		
		if(capability.getProperties().size() == 0) {
			return;
		}
		
		indent(code);
		code.append("private static final " 
			+ convertType(QName.class, classInfo) 
			+ "[] _PROPERTIES = new " 
			+ convertType(QName.class, classInfo) 
			+ "[]");
		newLine(code);
		addImport(QName.class);
		
		indent(code);
		generateOpenBlock(code);
		newLine(code);
		
		for(Iterator i=capability.getProperties().iterator(); i.hasNext();) {
			JavaProperty property = (JavaProperty)i.next();
			indent(2,code);
			code.append("new " 
				+ convertType(QName.class, classInfo) 
				+ "(NAMESPACE_URI, \"" 
				+ getPropertyName(property, false) 
				+ "\", PREFIX)");
			if(i.hasNext()) {
				code.append(",");
			}
			newLine(code);
		}
		
		indent(code);
		generateCloseBlock(code);
		statement(";",code);
		newLine(2,code);
		
		indent(code);
		code.append("public " 
			+ convertType(QName.class, classInfo) 
			+ "[] getPropertyNames()");
		newLine(code);
		
		indent(code);
		generateOpenBlock(code);
		newLine(code);
		
		indent(2,code);
		code.append("return _PROPERTIES;");
		
		newLine(code);
		indent(code);
		generateCloseBlock(code);
		newLine(2,code);
		
		for(Iterator i=capability.getProperties().iterator(); i.hasNext();) {
			JavaProperty property = (JavaProperty)i.next();
			indent(code);
			code.append("private " 
				+ convertType(property.getJavaType(), classInfo) 
				+ " _" 
				+ getPropertyName(property, false) 
				+ ";");
			newLine(2, code);
		}
	}
	
	protected void generateInitialize(ClassInfo classInfo, StringBuffer code) {				
		indent(code);
		code.append("public void initialize() throws SoapFault ");
		newLine(code);
		indent(code);
		generateOpenBlock(code);		
		addImport(SoapFault.class);
    
        generateSuperInitialize(code);

        newLine(code);
        
		if(classInfo.getCapability().isEmpty()) {
			generateEmptyInitTodo(code);
		} else {
			generatePropertyInitTodo(classInfo, code);
		}
		
		indent(code);
		generateCloseBlock(code);
		newLine(2, code);
	}

	private void generateSuperInitialize(StringBuffer code) {
		newLine(code);
		
		indent(2, code);
		code.append("//");
		newLine(code);
		
		indent(2, code);
		code.append("//" +_MESSAGES.get("InitializeComment", false));
		newLine(code);
		
		indent(2, code);
		code.append("//");
		newLine(code);
		
		indent(2, code);
		code.append("super.initialize();");		
		newLine(code);
	}

	protected void generateEmptyInitTodo(StringBuffer code) {
		indent(2, code);
		code.append("//");
		newLine(code);
		
		indent(2, code);		
		code.append("//" + _MESSAGES.get("InitEmptyTODO", false));
		newLine(code);
		
		indent(2, code);
		code.append("//");
		newLine(code);
	}

	protected void generatePropertyInitTodo(ClassInfo classInfo, StringBuffer code) {
		indent(2, code);
		code.append("//");
		newLine(code);

		indent(2, code);		
		code.append("//" + _MESSAGES.get("InitTODO", false));
		newLine(code);
		
		indent(2, code);
		code.append("//");
		newLine(code);
		
		for(Iterator i = classInfo.getCapability().getProperties().iterator(); i.hasNext(); ) {
			JavaProperty property = (JavaProperty)i.next();
			if(!property.getJavaType().isPrimitive()) {
				indent(2, code);
				code.append("// " + property.getName().getLocalPart());
				newLine(code);		
			}
		}
		
		indent(2, code);
		code.append("//");
		newLine(2, code);
		
		indent(2, code);		
		code.append("throw new RuntimeException(\"" + _MESSAGES.get("InitThrow", false) + "\");");
		newLine(code);
	}

	protected void generateGetOperationBody(JavaProperty property, StringBuffer code) {
		newLine(code);
		indent(code);
		generateOpenBlock(code);
		newLine(code);
		
		indent(2,code);
		statement("return _" 
			+ getPropertyName(property, false) 
			+  ";",code);
		newLine(code);
		
		indent(code);
		generateCloseBlock(code);
	}

	protected void generateOperationBody(JavaMethod method, StringBuffer code) {
		indent(code);
		generateOpenBlock(code);
		newLine(code);
		
		indent(2,code);
		code.append("//TODO implement " + getMethodName(method));
		newLine(code);
		
		indent(2,code);
		code.append("throw new RuntimeException(\"Unimplemented Method: " + getMethodName(method) + "\");");
		newLine(code);
		
		indent(code);
		generateCloseBlock(code);
	}

	protected void generateSetOperationBody(JavaProperty property, StringBuffer code) {
		newLine(code);
		indent(code);
		generateOpenBlock(code);
		newLine(code);
		
		indent(2,code);
		statement("_" 
			+ getPropertyName(property, false) 
			+ " = param0;",code);				
		newLine(code);
		
		indent(code);
		generateCloseBlock(code);	
	}
}
