/* 
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 *
 */

package org.apache.muse.tools.generator.util;

import java.io.File;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import org.apache.muse.util.messages.Messages;
import org.apache.muse.util.messages.MessagesFactory;
import org.apache.muse.ws.resource.metadata.MetadataDescriptor;
import org.w3c.dom.Document;

/**
 * Represents an object that is passed between phases of code generation. It 
 * contains the results of the phases of code generation as well as any component-specific
 * data. The reason this is so abstract is because the various pieces (projectizers, synthesizers,
 * analyzers) can all be so drastically different and can be replaced on a whim.
 * 
 * This class also contains static definitions of the parameter keys (this thing
 * is basically just wrapping a hash map) and metadata about the parameters. Again,
 * since we don't want to have these parameters pollute the code-generation APIs then
 * we need at least some sanity-checking. This is done with <code>checkConfiguration</code>
 * which takes a <code>Configurable</code> (which all code-generation components implement)
 * and a <code>ConfigurationData</code> and then makes sure the values passed in match
 * in class against what's expected. 
 * 
 * This isn't perfect but it's a decent start. The idea here is also to put in metadata
 * that is human-readable so that these objects can be used to populate option menus and
 * show command-line options dynamically.
 * 
 * @author Andrew Eberbach (aeberbac)
 */
public class ConfigurationData implements Cloneable {

	public static final String WSDL_DOCUMENT_LIST = "wsdl_documents";
	
	public static ConfigurationDataDescriptor WSDL_DOCUMENT_LIST_CONFIGURATION = 
		new ConfigurationDataDescriptor(
			WSDL_DOCUMENT_LIST,
			Document[].class);
		
	public static final String DESCRIPTOR_DOCUMENT = "descriptor_document";
	
	public static ConfigurationDataDescriptor DESCRIPTOR_DOCUMENT_CONFIGURATION = 
		new ConfigurationDataDescriptor(
				DESCRIPTOR_DOCUMENT,
				Document.class);
	
	public static final String OVERWRITE = "overwrite";
	
	public static final ConfigurationDataDescriptor OVERWRITE_CONFIGURATION =
		new ConfigurationDataDescriptor(
				OVERWRITE,
				Boolean.class);
	
	public static final String CAPABILITIES_MAP_LIST = "capabilities_map_list";
	
	public static final ConfigurationDataDescriptor CAPABILITIES_MAP_LIST_CONFIGURATION = 
		new ConfigurationDataDescriptor(
				CAPABILITIES_MAP_LIST,
				Map[].class);
	
	public static final String FILES_MAP_LIST = "files_map_list" ;
	
	public static final ConfigurationDataDescriptor FILES_MAP_LIST_CONFIGURATION =
		new ConfigurationDataDescriptor(
				FILES_MAP_LIST,
				Map[].class);

	public static final String TARGET_DIRECTORY = "target_dir";
	
	public static final ConfigurationDataDescriptor TARGET_DIRECTORY_CONFIGURATION =
		new ConfigurationDataDescriptor(
				TARGET_DIRECTORY,
				File.class);

	public static final String GENERATE_CUSTOM_HEADERS = "generate_custom_headers";
	
	public static final ConfigurationDataDescriptor GENERATE_CUSTOM_HEADERS_CONFIGURATION = 
		new ConfigurationDataDescriptor(
				GENERATE_CUSTOM_HEADERS,
				Boolean.class);

	public static final String IGNORE_SET_LIST = "ignore_sets";
	
	public static final ConfigurationDataDescriptor IGNORE_SET_LIST_CONFIGURATION = 
		new ConfigurationDataDescriptor(
				IGNORE_SET_LIST,
				Set[].class);

	public static final String METADATA_DESCRIPTOR_LIST = "metadata_list";
	
	public static final ConfigurationDataDescriptor METADATA_DESCRIPTOR_LIST_CONFIGURATION = 
		new ConfigurationDataDescriptor(
				METADATA_DESCRIPTOR_LIST,
				MetadataDescriptor[].class);
	
	private static Messages _MESSAGES = MessagesFactory.get(ConfigurationData.class);

	private Map _properties = new HashMap();
	
	public void addParameter(String parameter, Object value) {
		_properties.put(parameter, value);
	}
	
	public Object getParameter(String parameter) {
		return _properties.get(parameter);
	}

	/**
	 * This is a <b>shallow</b> clone of the map that backs this object.
	 * 
	 * @see java.lang.Object#clone()
	 */
	public Object clone() {
		ConfigurationData newData = new ConfigurationData();
		newData._properties = new HashMap(_properties);
		return newData;
	}		
		
	/**
	 * Given a <code>Configurable</code> and a <code>ConfigurationData</code> check to make sure
	 * that the <code>Configurable</code>'s required parameters are all present and are of the 
	 * correct class.
	 * 
	 * @param configurable
	 * 				The object whose configuration we're checking
	 * @param configuration
	 * 				The configuration we're checking
	 * 
	 * @throws Exception
	 */
	public static void checkConfiguration(Configurable configurable, ConfigurationData configuration) throws Exception {		
		if(configurable == null) {
			throw new IllegalArgumentException(_MESSAGES.get("NullConfigurable"));
		}
		
		if(configuration == null) {
			throw new IllegalArgumentException(_MESSAGES.get("NullConfiguration"));
		}
		
		ConfigurationDataDescriptor[] descriptors = 
			configurable.getConfigurationDataDescriptions();
		
		if(descriptors == null) {
			throw new IllegalArgumentException(_MESSAGES.get("NullConfiguration"));
		}
		
		for(int i=0; i < descriptors.length; i++) {
			ConfigurationDataDescriptor descriptor =
				descriptors[i];
			
			String key = descriptor.getParameterKey();
			Class parameterClass = descriptor.getParameterClass();
			
			Object value = configuration.getParameter(key);
			
			if(value == null) {
				Object[] filler = new Object[] {
					key	
				};
				throw new IllegalArgumentException(_MESSAGES.get("NullProperty", filler));
			}
			
			if( ! (parameterClass.isAssignableFrom(value.getClass())) ) {
				Object[] filler = new Object[] {
						parameterClass,
						value.getClass()
				};
				throw new IllegalArgumentException(_MESSAGES.get("InvalidType", filler));
			}
		}
	}

	public Map getProperties() {
		return _properties;
	}
}
