/* 
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 *
 */

package org.apache.muse.tools.generator.util;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;

import org.apache.muse.util.messages.Messages;
import org.apache.muse.util.messages.MessagesFactory;

public class OverwriteHelper {

	private static final String COMMENT_CHAR = "#";

	private static final String OVERWRITE_MANIFEST_FILE = ".overwrite";

	private static Messages _MESSAGES = MessagesFactory
			.get(OverwriteHelper.class);

	private Set _overwriteSet = null;

	private boolean _overwriteAll = false;

	private boolean _modifiable;

	public OverwriteHelper(File targetDirectory, boolean overwriteAll) {
		_overwriteSet = loadOverwriteSet(targetDirectory);
		_overwriteAll = overwriteAll;

		if (_overwriteSet == null) {
			_overwriteSet = new HashSet();
			_modifiable = true;
		} else {
			_modifiable = canOverwrite(new File(targetDirectory,
					OVERWRITE_MANIFEST_FILE), targetDirectory);
			if (_modifiable) {
				_overwriteAll = true;
			}
		}
	}

	public boolean canOverwrite(File file, File directory) {
		return (!file.exists() || _overwriteAll || overwriteSetContains(file,
				directory));
	}

	private boolean overwriteSetContains(File file, File directory) {
		return _overwriteSet.contains(getBaseFile(file, directory));
	}

	private File getBaseFile(File file, File directory) {
		try {
			String basePath = directory.getCanonicalPath();
			String filePath = file.getCanonicalPath();
			return new File(filePath.substring(basePath.length() + 1));
		} catch (Exception e) {
			throw new RuntimeException();
		}
	}

	/**
	 * Load an overwrite set manifest from a given directory. The
	 * 
	 * @param targetDirectory
	 * @param initialSet
	 */
	private Set loadOverwriteSet(File targetDirectory) {
		File overwriteManifest = new File(targetDirectory,
				OVERWRITE_MANIFEST_FILE);

		if (!overwriteManifest.exists()) {
			return null;
		}

		Set overwriteSet = new HashSet();

		try {
			BufferedReader br = new BufferedReader(new FileReader(
					overwriteManifest));

			String line = null;
			while ((line = br.readLine()) != null) {
				if (!line.startsWith(COMMENT_CHAR)) {
					overwriteSet.add(new File(line));
				}
			}

			br.close();
		} catch (IOException e) {
			throw new RuntimeException(
					_MESSAGES.get("FailedOverwriteManifest"), e);
		}

		return overwriteSet;
	}

	public void writeToDirectory(File directory) throws Exception {
		File overwriteManifest = new File(directory, OVERWRITE_MANIFEST_FILE);
		if (canOverwrite(overwriteManifest, directory)) {
			BufferedWriter bw = new BufferedWriter(new FileWriter(overwriteManifest));
			
			Object[] files = _overwriteSet.toArray();
			Arrays.sort(files);
			
			for (int i=0; i < files.length; i++) {
				bw.write(files[i].toString() + "\n");
			}
			
			bw.close();
		}
	}

	public void add(File file, File directory) {
		if (_modifiable) {
			_overwriteSet.add(getBaseFile(file, directory));
		}
	}
}
