/* 
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 *
 */

package org.apache.muse.tools.generator.util;

import java.io.InputStream;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.apache.muse.util.FileUtils;
import org.apache.muse.util.xml.XmlUtils;
import org.apache.muse.ws.notification.WsnConstants;
import org.apache.muse.ws.notification.impl.SimpleSubscriptionManager;
import org.apache.muse.ws.resource.lifetime.WsrlConstants;
import org.apache.muse.ws.resource.lifetime.impl.SimpleImmediateTermination;
import org.apache.muse.ws.resource.lifetime.impl.SimpleScheduledTermination;
import org.apache.muse.ws.resource.metadata.MetadataDescriptor;
import org.apache.muse.ws.resource.properties.WsrpConstants;
import org.apache.muse.ws.resource.properties.get.impl.SimpleGetCapability;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

/**
 * Endpoints that have <code>resource-type</code>s that are <code>NotificationProducer</code>s
 * need to have a <code>SubscriptionManager</code> included as a sibling <code>resource-type</code>. 
 * Without this <code>SubscriptionManager</code> the <code>NotificationProducer</code> capability
 * will not function. This class will analyze the <code>ConfigurationData</code> and modify it
 * if a <code>resource-type</code> that has a <code>SubscriptionManager</code> is not found. 
 * 
 * @author Andrew Eberbach (aeberbac)
 */
public class SubscriptionManagerHelper {
	
    public static final String SUBSCRIPTION_MANAGER_DIR = "/resources/common/subscription-manager";
    public static final String SUBSCRIPTION_MANAGER_RESOURCE_TYPE = SUBSCRIPTION_MANAGER_DIR + "/muse.xml";
    public static final String SUBSCRIPTION_MANAGER_WSDL = SUBSCRIPTION_MANAGER_DIR + "/SubscriptionManager.wsdl";
		
	public static ConfigurationData addSubscriptionManager(ConfigurationData data) {
		
		if(needsSubscriptionManager(data)) {
			modifyDescriptorDocument(data);
            addSubscriptionManagerCapabilities(data);
			addSubscriptionManagerWsdl(data);
			addSubscriptionManagerMetadata(data);
		}
		
		return data;
	}
    
    private static void addSubscriptionManagerMetadata(ConfigurationData data) {
		MetadataDescriptor[] oldMetadatas = (MetadataDescriptor[])data.getParameter(ConfigurationData.METADATA_DESCRIPTOR_LIST);
		MetadataDescriptor[] newMetadatas = new MetadataDescriptor[oldMetadatas.length + 1];
		System.arraycopy(oldMetadatas, 0, newMetadatas, 0, oldMetadatas.length);
		
		//
		// Explicitly set the last metadata to null since there
		// is no metadata associated with the subscription manager we're generating
		//
		newMetadatas[oldMetadatas.length] = null;
		
		data.addParameter(ConfigurationData.METADATA_DESCRIPTOR_LIST, newMetadatas);
	}

	private static void addSubscriptionManagerCapabilities(ConfigurationData data){
        Map[] oldCaps = (Map[])data.getParameter(ConfigurationData.CAPABILITIES_MAP_LIST);
        Map[] newCaps = new Map[oldCaps.length + 1];
        System.arraycopy(oldCaps, 0, newCaps, 0, oldCaps.length);
        
        Map subMgrCaps = new HashMap();
        
        Capability wsrpGet = new Capability(WsrpConstants.GET_CAPABILITY);
        wsrpGet.setBuiltIn(true);
        wsrpGet.setImplementingClass(SimpleGetCapability.class.getName());
        
        Capability wsrlImmediate = new Capability(WsrlConstants.IMMEDIATE_TERMINATION_URI);
        wsrlImmediate.setBuiltIn(true);
        wsrlImmediate.setImplementingClass(SimpleImmediateTermination.class.getName());
        
        Capability wsrlScheduled = new Capability(WsrlConstants.SCHEDULED_TERMINATION_URI);
        wsrlScheduled.setBuiltIn(true);
        wsrlScheduled.setImplementingClass(SimpleScheduledTermination.class.getName());
        
        Capability subMgr = new Capability(WsnConstants.SUBSCRIPTION_MGR_URI);
        subMgr.setBuiltIn(true);
        subMgr.setImplementingClass(SimpleSubscriptionManager.class.getName());
        
        subMgrCaps.put(wsrpGet.getURI(), wsrpGet);
        subMgrCaps.put(wsrlImmediate.getURI(), wsrlImmediate);
        subMgrCaps.put(wsrlScheduled.getURI(), wsrlScheduled);
        subMgrCaps.put(subMgr.getURI(), subMgr);
        
        newCaps[oldCaps.length] = subMgrCaps;
        
        data.addParameter(ConfigurationData.CAPABILITIES_MAP_LIST, newCaps);
    }
	
	private static void addSubscriptionManagerWsdl(ConfigurationData data) {        
		Document[] oldWsdls = (Document[]) data.getParameter(ConfigurationData.WSDL_DOCUMENT_LIST);
		Document[] newWsdls = new Document[oldWsdls.length + 1];
		System.arraycopy(oldWsdls, 0, newWsdls, 0, oldWsdls.length);
		newWsdls[oldWsdls.length] = loadDocument(SUBSCRIPTION_MANAGER_WSDL);
        
        data.addParameter(ConfigurationData.WSDL_DOCUMENT_LIST, newWsdls);
	}

	private static void modifyDescriptorDocument(ConfigurationData data) {
		Document subscriptionType = loadDocument(SUBSCRIPTION_MANAGER_RESOURCE_TYPE);
		
		Document descriptorDocument = (Document)data.getParameter(ConfigurationData.DESCRIPTOR_DOCUMENT);
		
		Node importedSubscriptionType = descriptorDocument.importNode(subscriptionType.getDocumentElement(), true);
		
		insertAfterLastResourceType(descriptorDocument, importedSubscriptionType);
	}
	
	private static void insertAfterLastResourceType(Document descriptorDocument, Node importedSubscriptionType) {		
		Element nodeAfterLastResourceType = getNodeAfterLastResourceType(descriptorDocument);
		
		if(nodeAfterLastResourceType == null) {
			descriptorDocument.getDocumentElement().appendChild(importedSubscriptionType);
		} else {
			descriptorDocument.getDocumentElement().insertBefore(nodeAfterLastResourceType, importedSubscriptionType);
		}
	}

	private static Element getNodeAfterLastResourceType(Document descriptorDocument) {
		Element[] elements = XmlUtils.getAllElements(descriptorDocument.getDocumentElement());
		
		//
		// This is the marker for the last resource-type element. We're going to do this by finding
		// the last resource-type element and then repeatedly doing a getNextSibling, 
		// checking to see if it's an Element until it is, or the getNextSibling is null.
		//
		// Remember, getNextSibling returns a *Node* so who knows if it's not a text node...
		//
		Node last = elements[elements.length - 1].getNextSibling();
		
		while(last != null) {	
			
			//
			// If this is true it has to be a non-resource-type element
			//
			if(last instanceof Element) {
				return (Element)last;
			}
			last = last.getNextSibling();
		} 
		
		//
		// We ran off the end
		//
		return null;
	}

	private static Document loadDocument(String path) {
		InputStream inputStream = FileUtils.loadFromContext(SubscriptionManagerHelper.class, path);
		try {
			return XmlUtils.createDocument(inputStream);
		} catch (Exception e) {
			throw new RuntimeException(e.getMessage(), e);
		}
	}

	private static boolean needsSubscriptionManager(ConfigurationData data) {
		Map[] capabilityMaps = (Map[])data.getParameter(ConfigurationData.CAPABILITIES_MAP_LIST);
		
		boolean hasNotificationProducer = false;
		boolean hasSubscriptionManager = false;		
		
		for(int j=0; j < capabilityMaps.length; j++) {
			for(Iterator i = capabilityMaps[j].values().iterator(); i.hasNext(); ) {
				Capability capability = (Capability)i.next();				
				String uri = capability.getURI();
				if(uri.equals(WsnConstants.SUBSCRIPTION_MGR_URI)) {
					hasSubscriptionManager = true;
				}
				if(uri.equals(WsnConstants.PRODUCER_URI)) {
					hasNotificationProducer = true;
				}
			}
		}
		
		return hasNotificationProducer && !hasSubscriptionManager;
	}
}
